
/*
    opti9xx.c - ALSA driver for OPTi 82C9xx based soundcards.
    Copyright (C) 1998-99 by Massimo Piccioni <piccio@caronte.csr.unibo.it>

    This code has been developed at the Italian Ministry of Air Defence,
    Sixth Division (oh, che gioia ...), Rome.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/*
    This code provides support for most OPTi 82C9xx chips based soundcards.
    All non-PnP chips (i.e. 82C928, 82C929 and 82C930) are fully supported.
    All PnP-capable chips (82C924, 82C925, 82C931 and 82C933) should be
    supported in non-PnP mode. The PnP support is still uncertain (see the
    PnP considerations below).

    This code is not complete! I still have to do some work (see the todo
    section below).

    This driver supports only one soundcard per system, due to an harware
    limitation (i/o at fixed locations).

    This driver allow the user to enable and fully configure the on-board
    cdrom drive interface (i.e. IDE/ATAPI, Sony, Mitsumi and Panasonic)
    of several OPTi based soundcards.
    Search for the NON-SCSI/ATAPI cdrom support documentation in the Linux
    kernel sources.

    Sources:
        OPTi Inc. 82C9xx databooks (available from ftp.opti.com).
        Analog Devices Inc. AD1845/8 databooks.
        Crystal Semiconductor Corp. CS4231A databook.
        Advanced Linux Sound Architecture documentation and sources.

    Version history:
        0.1:  82C9xx detection and configuration only
        1.0:  basic sound support added (ad1848 and opl3)
        1.1:  cdrom and game port configuration added
        1.1c: cdrom and game port configuration improved
        1.2:  most code rewritten, cs4231 support added
        1.3:  mpu-401 support added (82C929+ chips only)

    Todo:
        add PnP support (ALSA native)
*/

/*
    Chips description coming soon ...
*/

/*
    PnP considerations:
        At the moment, this driver is not reliable with PnP-capable chips,
        because the MC registers configuration doesn't work very well when
        PnP is enabled.
        Several soundcards are PnP-only, others have a jumper that enables
        or disables PnP. If you can, disable PnP, if not, don't worry, I'm
        working on PnP support ;)

        With a non-PnP BIOS any OPTi PnP soundcard will not work, until a
        PnP activation procedure is applied (all logical devices will remain
        inactive after power-on).
        At the moment, you will need to use isapnptools to configure and
        activate your PnP card.

        If you don't have isapnptools already installed on your system, try
        at the following web site: http://www.roestock.demon.co.uk/isapnptools
*/

/*
    Sound cards (being) tested:
        Audio 16 Pro  EPC-SOUN9301      (82C930 based)
        ExpertColor  MED-3931 v2.0      (82C931 based)
        ExpertMedia Sound 16  MED-1600  (82C928 based - AD1848)
        Mozart  S601206-G               (OTI601 based - CS4231)
        Sound Player S-928              (82C928 based - AD1848)

    Cdrom drives tested:
        CreativeLabs CD200F (Funai E2550UA), Panasonic i/f
        Matsushita-Kotobuki CR-562-B, Panasonic i/f

    Things of interest:
        Several soundcards (such as ExpertMedia Sound 16) use jumpers to
        configure the on-board cdrom drive interface, so software configuration
        doesn't work (of course).

        Be careful when configuring the on-board cdrom drive interface as IDE,
        because the system may hang (i/o and irq conflicts).

        I have no problems with my CreativeLabs CD200F cdrom drive (a Funai
        E2550UA drive with Panasonic proprietary interface) connected to the
        on-board interface of a S-928 soundcard (see the cdrom configuration
        section below). I simply load the sbpcd kernel module after this.
        See the sbpcd documentation for more informations (available in the
        /usr/src/linux/Documentation/cdrom directory).

        Another interesting thing: the contemporary presence of an 82C928
        based (not PnP) sound card and an IBM Auto 16/4 Token Ring ISA PnP
        adapter, may cause several problems (i.e. Linux might not detect the
        token ring adapter), due to an i/o conflict between the soundcard
        game port (0xa20 fixed?) and the tr adapter base port (0xa20 or 0xa24).
        It happens because PnP BIOS is not able to detect what i/o ports the
        soundcard is using, so the first 'available' i/o port (i.e. 0xa20) is
        assigned to the tr adapter (i/o conflict).
        A fine solution is to use the isapnptools to reconfigure your PnP
        adapters (then you can load the ibmtr kernel module without any
        problem).
        Another solution is to disable the sound card game port in this module
        (loaded before the ibmtr module, see the game port configuration
        section below).

        Finally, Mozart soundcards are also supported, due to some kind of
        compatibility between OPTi 82C928 and OAK OTI601 chips (see mozart.c
        in the ALSA sources).
        I was not able to get the on-board cdrom interface working with a
        Mozart soundcard (broken board?). Unfortunately I have no documentation
        about the OTI601 chip, so any suggestion is welcome.
*/

/*
    Your experiences may be very useful (there is a lot of OPTi 82C9xx based
    soundcards), so feel free to contact me by email.

    Thanks to:
        Elliot Lee, the 82C930 based soundcards pioneer.
*/

/*
    Module options:
        snd_port        codec base i/o port: 0x530,0xe80,0xf40,0x604
        snd_irq         WSS irq num: 5*,7,9,10,11 (*: 82C930+ only) 
        snd_dma1        WSS dma1 num: 0,1,3
        snd_dma2        WSS dma2 num: 0,1
        snd_dma1_size   dma1 size in kB: 8,16,32,64
        snd_dma2_size   dma2 size in kB: 8,16,32,64
        snd_mpu_port    mpu-401 base i/o port: 0x300,0x310,0x320,0x330
        snd_mpu_irq     mpu-401 irq num: 5,7,9,10
        cd_if           cdrom interface: disabled,ide1,ide2,mitsumi,panasonic,
                        sony
        cd_port         cdrom i/o port: 0x320,0x330,0x340,0x360
        cd_irq          cdrom irq num: -1,5,7,9,10,11          (-1 disables)
        cd_drq          cdrom drq num: -1,0,1,3                (-1 disables)
        game            game port: enabled,disabled

    Default values:
        snd_xxx         auto (probed and configured)
        cd_if           disabled
        cd_port         0x340
        cd_irq          disabled
        cd_drq          disabled
        game            disabled

    Examples:
        modprobe snd-opti9xx snd_irq=9
        modprobe snd-opti9xx cd_port=0x360 game=enabled snd_port=0xe80
        modprobe snd-opti9xx snd_port=0xf40 snd_irq=9 snd_dma2=1
        modprobe snd-opti9xx snd_mpu_port=0x310 snd_irq=10
*/

#define OPTi9XX_CURRENT_VERSION		1.3

#define OPTi9XX_DEBUG
#undef OPTi9XX_DEBUG_MC
#define OPTi9XX_EXPERIMENTAL

#define __SND_OSS_COMPAT	/* why not? */
#define SND_MAIN_OBJECT_FILE
#include "driver.h"
#include "ad1848.h"
#include "cs4231.h"
#include "initval.h"
#include "isapnp.h"
#include "mpu401.h"
#include "opl3.h"

#ifndef ESUCCESS
#define ESUCCESS	0
#endif	/* ESUCCESS */

#define opti9xx_printk( args... )	snd_printk( "opti9xx: " ##args )

/* on-board interfaces configuration section */
/* cdrom type selection */
static struct {
	char *name;
	unsigned char bits;
} cd_if_table[]= {
	{ "disabled",	0x00 },		/* all */
	{ "ide1",	0x05 },		/* c930 */
	{ "ide2",	0x04 },		/* c924, c925, c930, c931 */
	{ "mitsumi",	0x02 },		/* c924, c928, c929 */
	{ "panasonic",	0x03 },		/* c924, c928, c929 */
	{ "sony",	0x01 }		/* c924, c928, c929 */
};

/* cdrom ioports (Sony, Mitsumi or Panasonic interfaces only) */
static struct {
	int port;
	unsigned char bits;
} cd_port_table[]= {
	{ 0x320,	0x03 },		
	{ 0x330,	0x01 },
	{ 0x340,	0x00 },		
	{ 0x360,	0x02 }
};

/* cdrom irq (Sony, Mitsumi, Panasonic) */
static struct {
	int irq;
	unsigned char bits;
} cd_irq_table[]= {
	{ -1,	0x00 },		/* disabled */
	{ 5,	0x01 },
	{ 7,	0x02 },
	{ 9,	0x04 },
	{ 10,	0x05 },
	{ 11,	0x06 }
};

/* cdrom drq (Sony, Mitsumi, Panasonic) */
static struct {
	int drq;
	unsigned char bits;
} cd_drq_table[]= {
	{ -1,	0x03 },		/* disabled */
	{ 0,	0x01 },
	{ 1,	0x02 },
	{ 3,	0x00 }
};

/* game port */
static struct {
	char *name;
	unsigned char bits;
} game_en_table[]= {
	{ "enabled",	0x00 },	
	{ "disabled",	0x01 }
};

#define OPTi9XX_OPL3_LPORT	0x388
#define OPTi9XX_OPL3_RPORT	OPTi9XX_OPL3_LPORT+2

/* OPTi chips supported */
enum chipsets {
	c928=0,	c929=1,	c924=2,	c925=3,
	c930=4,	c931=5,	c933=6,	unsupported=c933
};

/* chip-specific parameters */
static struct {
	char *name;
	unsigned char pwd;
	int mc_base;
	char mc_regs;
	char mc_size;
	int pwd_reg;
} chip_info[]= {
	{ "82C928",		0xe2,	0xf8c,	6,	7,	3 },
	{ "82C929",		0xe3,	0xf8c,	6,	7,	3 },
	{ "82C924",		0xe5,	0xf8c,	9,	10,	3 },
	{ "82C925",		0xe5,	0xf8c,	9,	10,	3 },
	{ "82C930",		0xe4,	0xf8f,	12,	1,	0 },
	{ "82C931",		0xe4,	0xf8d,	26,	1,	0 },
	{ "unsupported",	0x00,	0x000,	0,	0,	0 }
};

#define OPTi9XX_CHIP		chip_info[opti9xx_chip].name
#define OPTi9XX_MC_BASE		chip_info[opti9xx_chip].mc_base
#define OPTi9XX_MC_REGS		chip_info[opti9xx_chip].mc_regs
#define OPTi9XX_MC_SIZE		chip_info[opti9xx_chip].mc_size
#define OPTi9XX_PASSWD		chip_info[opti9xx_chip].pwd

#define OPTi9XX_MC(x)		x
#define OPTi92X_MC_REG(x)	OPTi9XX_MC_BASE+OPTi9XX_MC(x)
#define OPTi9XX_PWD_REG		OPTi9XX_MC_BASE+chip_info[opti9xx_chip].pwd_reg

#define OPTi924_MC_INDEX	OPTi92X_MC_REG(8)
#define OPTi924_MC_DATA		OPTi92X_MC_REG(9)
#define OPTi924_MC_DIR_SIZE	7

#define OPTi93X_MC_INDEX	opti93x_mc_index
#define OPTi93X_MC_DATA		OPTi93X_MC_INDEX+1
#define OPTi93X_MC_INDIR_SIZE	2

#define OPTi93X_ENABLE_PWD	0x00
#define OPTi9XX_ENABLE_MPU	0x80
#define OPTi9XX_ENABLE_WSS	0x80
#define OPTi93X_IDX_BITS	((OPTi93X_MC_INDEX & (1 << 8)) >> 4)|\
				((OPTi93X_MC_INDEX & (0xf0)) >> 4)

#define WSS_CONFIG_PORT		snd_port
#define WSS_CONFIG_SIZE		4
#define WSS_BASE_PORT		WSS_CONFIG_PORT+WSS_CONFIG_SIZE
#define WSS_STATUS_PORT		WSS_BASE_PORT+2
#define WSS_SIZE		4

#define MPU401_TOTAL_SIZE	2
#define OPL3_TOTAL_SIZE		4
#define WSS_TOTAL_SIZE		WSS_SIZE+WSS_CONFIG_SIZE

static int wss_dma1_bits[] = {
	1,	2,	0,	3
};

static int wss_irq_bits[] = {
	0,	0,	0,	0,	0,	5,
	0,	1,	0,	2,	3,	4
};

static int wss_valid_ports[]= {
	0x530,	0xe80,	0xf40,	0x604,	SND_AUTO_PORT
};

static int wss_valid_dma1[]= {
	3,	1,	0,	-1
};

static int wss_valid_dma2[][2]= {
	{ 1,	-1 },
	{ 0,	-1 },
	{ -1, 	-1 },
	{ 0,	-1 }
};

static int wss_valid_irqs[]= {
	5,	/* 82C93x based soundcards only */
	9,	10,	11,
	7,	/* parallel port, problems with several mother-boards */
	-1
};

static int mpu_irq_bits[] = {
	0,	0,	0,	0,	0,	2,
	0,	3,	0,	0,	1
};

static int mpu_valid_ports[]= {
	0x330,	0x320,	0x310,	0x300,	SND_AUTO_PORT
};

static int mpu_valid_irqs[]= {
	5,	9,	10,
	7,	/* parallel port, problems with several mother-boards */
	-1
};

/* WSS configuration (module options) */
int snd_dma1 = SND_DEFAULT_DMA1;
int snd_dma2 = SND_DEFAULT_DMA1;
int snd_dma1_size = SND_DEFAULT_DMA_SIZE1;
int snd_dma2_size = SND_DEFAULT_DMA_SIZE1;
char *snd_id = SND_DEFAULT_STR1;
int snd_index = SND_DEFAULT_IDX1;
int snd_irq = SND_DEFAULT_IRQ1;
int snd_port=SND_DEFAULT_PORT1;
int snd_mpu_irq = SND_DEFAULT_IRQ1;
int snd_mpu_port = SND_DEFAULT_PORT1;

/* cdrom drive and game port configuration section (module options) */
int cd_drq = -1;
char *cd_if = "disabled";
int cd_irq = -1;
int cd_port = 0x340;
/* default: game port disabled (i/o conflict with an ibmtr adapter) */
char *game = "disabled";

#ifdef MODULE_PARM
MODULE_AUTHOR("Massimo Piccioni <piccio@caronte.csr.unibo.it>");
MODULE_DESCRIPTION("OPTi 82C9xx based soundcards support.");
/* cdrom interface configuration */
MODULE_PARM(cd_drq, "i");
MODULE_PARM_DESC(cd_drq, "cdrom drive dma number");
MODULE_PARM(cd_irq, "i");
MODULE_PARM_DESC(cd_irq, "cdrom drive irq number");
MODULE_PARM(cd_port, "i");
MODULE_PARM_DESC(cd_port, "cdrom drive interface port");
MODULE_PARM(cd_if, "s");
MODULE_PARM_DESC(cd_if, "cdrom drive interface type");
/* game port configuration */
MODULE_PARM(game, "s");
MODULE_PARM_DESC(game, "game port enable");
/* WSS configuration */
MODULE_PARM(snd_dma1, "i");
MODULE_PARM_DESC(snd_dma1, "WSS dma1 number");
MODULE_PARM(snd_dma1_size, "i");
MODULE_PARM_DESC(snd_dma1_size, "WSS dma1 size in kB");
MODULE_PARM(snd_dma2, "i");
MODULE_PARM_DESC(snd_dma2, "WSS dma2 number");
MODULE_PARM(snd_dma2_size, "i");
MODULE_PARM_DESC(snd_dma2_size, "WSS dma2 size in kB");
MODULE_PARM(snd_irq, "i");
MODULE_PARM_DESC(snd_irq, "WSS irq number");
MODULE_PARM(snd_port, "i");
MODULE_PARM_DESC(snd_port, "sound base i/o port");
/* MPU-401 configuration */
MODULE_PARM(snd_mpu_irq, "i");
MODULE_PARM_DESC(snd_mpu_irq, "MPU-401 irq number");
MODULE_PARM(snd_mpu_port, "i");
MODULE_PARM_DESC(snd_mpu_port, "MPU-401 base i/o port");
#endif	/* MODULE_PARM */

/* internal data */
snd_irq_t *snd_mpu_irq_ptr = NULL;
snd_card_t *opti9xx_card = NULL;
snd_dma_t *snd_dma1_ptr = NULL;
snd_dma_t *snd_dma2_ptr = NULL;
snd_irq_t *snd_irq_ptr = NULL;
snd_rawmidi_t *snd_midi = NULL;
snd_kmixer_t *snd_mixer = NULL;
snd_synth_t *snd_opl3 = NULL;
snd_pcm_t *snd_pcm = NULL;

unsigned char cd_drq_sel = 0xff;
unsigned char cd_if_sel = 0xff;
unsigned char cd_irq_sel = 0xff;
unsigned int cd_port_sel = 0xff;
unsigned char game_sel = 0xff;

int opti9xx_chip = unsupported;
struct isapnp_dev *opti9xx_isapnp_dev = NULL;
int opti93x_mc_index = 0xe0e;		/* 0x[ef][0..f]e are valid values */ 
void (*opti9xx_wss_irq_ptr)(snd_pcm_t *pcm, unsigned char status) = NULL;


void snd_opti9xx_use_inc(snd_card_t *card)
{
	MOD_INC_USE_COUNT;	/* wow, someone is using this driver! */
}

void snd_opti9xx_use_dec(snd_card_t *card)
{
	MOD_DEC_USE_COUNT;	/* see you soon ;) */
}


/* mpu irq handling routine */
void snd_opti9xx_mpu_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
	snd_mpu401_uart_interrupt(snd_midi);
}

/* wss irq handling routine, see card-ad1848.c or card-cs4231.c */
void snd_opti9xx_wss_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
	register unsigned char status = inb(WSS_STATUS_PORT);
	if (status & 0x01)
		opti9xx_wss_irq_ptr(snd_pcm, status);
}


unsigned char snd_opti9xx_read(unsigned int idx)
{
	unsigned long flags;
	unsigned char retval = 0xff;
	if (idx > OPTi9XX_MC_REGS)
		return retval;
	snd_cli(&flags);
	outb(OPTi9XX_PASSWD, OPTi9XX_PWD_REG);
	switch (opti9xx_chip) {
	case c924:
	case c925:
		if (idx > OPTi924_MC_DIR_SIZE) {
			outb(idx, OPTi924_MC_INDEX);
			outb(OPTi9XX_PASSWD, OPTi9XX_PWD_REG);
			retval = inb(OPTi924_MC_DATA);
			break;
		}
	case c928:
	case c929:
		retval = inb(OPTi92X_MC_REG(idx));
		break;
	case c930:
	case c931:
		outb(idx, OPTi93X_MC_INDEX);
		outb(OPTi9XX_PASSWD, OPTi9XX_PWD_REG);
		retval = inb(OPTi93X_MC_DATA);
		break;
	default:	/* uff */
	}
	snd_sti(&flags);
#ifdef OPTi9XX_DEBUG_MC
	opti9xx_printk("0x%x read from MC%d reg\n", retval, idx);
#endif	/* OPTi9XX_DEBUG_MC */
	return retval;
}

void snd_opti9xx_write(unsigned int idx, unsigned char value)
{
	unsigned long flags;
	if (idx > OPTi9XX_MC_REGS)
		return;
	snd_cli(&flags);
	outb(OPTi9XX_PASSWD, OPTi9XX_PWD_REG);
	switch (opti9xx_chip) {
	case c924:
	case c925:
		if (idx > OPTi924_MC_DIR_SIZE) {
			outb(idx, OPTi924_MC_INDEX);
			outb(OPTi9XX_PASSWD, OPTi9XX_PWD_REG);
			outb(value, OPTi924_MC_DATA);
			break;
		}
	case c928:
	case c929:
		outb(value, OPTi92X_MC_REG(idx));
		break;
	case c930:
	case c931:
		outb(idx, OPTi93X_MC_INDEX);
		outb(OPTi9XX_PASSWD, OPTi9XX_PWD_REG);
		outb(value, OPTi93X_MC_DATA);
		break;
	default:	/* uff */
	}
	snd_sti(&flags);
#ifdef OPTi9XX_DEBUG_MC
	opti9xx_printk("written 0x%x to MC%d reg\n", value, idx);
#endif	/* OPTi9XX_DEBUG_MC */
}

void snd_opti9xx_conf(unsigned int idx, unsigned char value, unsigned char mask)
{
	unsigned char old = snd_opti9xx_read(idx);
	unsigned char new = (old & ~mask) | (value & mask);
	snd_opti9xx_write(idx, new);
}

#ifdef OPTi9XX_DEBUG
void snd_opti9xx_dump(void)
{
	int i;
	opti9xx_printk("dumping mc registers: ");
	for (i = 1; i <= OPTi9XX_MC_REGS; i++)
		printk("0x%2.2x ", snd_opti9xx_read(i));
	printk("\n");
}
#endif	/* OPTi9XX_DEBUG */


int snd_opti9xx_config_enable_wss(void)
{
	if (opti9xx_chip < c930)
		snd_opti9xx_conf(OPTi9XX_MC(1), OPTi9XX_ENABLE_WSS, 0x80);
	else
		snd_opti9xx_conf(OPTi9XX_MC(6), OPTi9XX_ENABLE_WSS >> 6, 0x03);
	return ESUCCESS;
}

int snd_opti9xx_config_onboard_if(void)
{
	snd_opti9xx_conf(OPTi9XX_MC(1), (cd_if_sel << 1) | game_sel, 0x0f);
	if (opti9xx_chip < c925)
		snd_opti9xx_conf(OPTi9XX_MC(2),
			(cd_port_sel << 6) | (cd_irq_sel << 2) | cd_drq_sel,
			0xdf);
	return ESUCCESS;
}

int snd_opti9xx_config_isa(void)
{
	int error;
	if ((error = snd_opti9xx_config_enable_wss()))
		return error;
	if ((error = snd_opti9xx_config_onboard_if()))
		return error;
	switch (opti9xx_chip) {
	case c924:
	case c925:
		snd_opti9xx_conf(OPTi9XX_MC(2), 0x00, 0x20);
		snd_opti9xx_conf(OPTi9XX_MC(3), 0xf0, 0xff);
		snd_opti9xx_conf(OPTi9XX_MC(4), 0x00, 0x0c);
		break;
	case c928:
	case c929:
		snd_opti9xx_conf(OPTi9XX_MC(2), 0x00, 0x20);
		snd_opti9xx_conf(OPTi9XX_MC(3), 0xf2, 0xfe);
		snd_opti9xx_conf(OPTi9XX_MC(4), 0x00, 0x0c);
		break;
	case c930:
	case c931:
		snd_opti9xx_conf(OPTi9XX_MC(3), 0x00, 0xff);
		snd_opti9xx_conf(OPTi9XX_MC(4), 0x10, 0x3c);
		snd_opti9xx_conf(OPTi9XX_MC(5), 0x00, 0xbf);
		break;
	}
	return ESUCCESS;
}

int snd_opti9xx_config_mpu(unsigned char mpu_base_bits)
{
	snd_opti9xx_conf(OPTi9XX_MC(6),
		OPTi9XX_ENABLE_MPU |
		(mpu_base_bits << 5) |
		(mpu_irq_bits[snd_mpu_irq] << 3), 0xf8);
	return ESUCCESS;
}

int snd_opti9xx_config_pnp(void)
{
	return -ENOSYS;
}

int snd_opti9xx_config_wss_base(unsigned char wss_base_bits)
{
	snd_opti9xx_conf(OPTi9XX_MC(1), wss_base_bits << 4, 0x30);
	return ESUCCESS;
}

int snd_opti9xx_config_wss_dmas_irq(void)
{
	unsigned char conf;
	conf = wss_irq_bits[snd_irq] << 3;
	conf |= wss_dma1_bits[snd_dma1];
	if (snd_dma2_ptr)
		conf |= 0x04;
	outb(conf, WSS_CONFIG_PORT);
	return ESUCCESS;
}


int snd_opti93x_resources_direct_regs(void)
{
	int hb, lb;
	for (hb = 0x00; hb < 0x02; hb++)
		for (lb = 0x00; lb < 0x10; lb++) {
			opti93x_mc_index = 0xe0e + (hb << 8) + (lb << 4);
			if (!snd_register_ioport(opti9xx_card,
				OPTi93X_MC_INDEX, OPTi93X_MC_INDIR_SIZE,
				"opti93x - direct regs", NULL))
				return ESUCCESS;
		}
	return -EBUSY;
}

int snd_opti9xx_resources_mc_regs(void)
{
	return snd_register_ioport(opti9xx_card,
			OPTi9XX_MC_BASE, OPTi9XX_MC_SIZE,
			"opti9xx - mc regs", NULL);
}

int snd_opti9xx_resources_mpu401(void)
{
	int bits, error;
	if (snd_mpu_port != SND_AUTO_PORT)
		if (!snd_register_ioport(opti9xx_card,
			snd_mpu_port, MPU401_TOTAL_SIZE,
			"opti9xx - mpu401",
			NULL))
			snd_mpu_port = SND_AUTO_PORT;
	for (bits = 0x00; mpu_valid_ports[bits] != SND_AUTO_PORT; bits++)
		if (snd_mpu_port == SND_AUTO_PORT)
			if (!snd_register_ioport(opti9xx_card,
				mpu_valid_ports[bits], MPU401_TOTAL_SIZE,
				"opti9xx - mpu401",
				NULL)) {
				snd_mpu_port = mpu_valid_ports[bits];
				break;
			} else
				continue;
		else if (snd_mpu_port == mpu_valid_ports[bits])
			break;
	if (snd_mpu_port == SND_AUTO_PORT)
		return -EBUSY;
	if ((error = snd_register_interrupt(opti9xx_card,
		"opti9xx - mpu401",
		snd_mpu_irq, SND_IRQ_TYPE_ISA, snd_opti9xx_mpu_interrupt,
		NULL,
		mpu_valid_irqs,
		&snd_mpu_irq_ptr)))
		return error;
	snd_mpu_irq = snd_mpu_irq_ptr->irq;
	return snd_opti9xx_config_mpu(bits);
}

int snd_opti9xx_resources_opl3(void)
{
	return snd_register_ioport(opti9xx_card,
		OPTi9XX_OPL3_LPORT, OPL3_TOTAL_SIZE,
		"opti9xx - opl3", NULL);
}

int snd_opti9xx_resources_wss_base(void)
{
	int bits;
	if (snd_port != SND_AUTO_PORT)
		if (!snd_register_ioport(opti9xx_card,
			snd_port, WSS_TOTAL_SIZE,
			"opti9xx - wss",
			NULL))
			snd_port = SND_AUTO_PORT;
	for (bits = 0x00; wss_valid_ports[bits] != SND_AUTO_PORT; bits++)
		if (snd_port == SND_AUTO_PORT)
			if (!snd_register_ioport(opti9xx_card,
				wss_valid_ports[bits], WSS_TOTAL_SIZE,
				"opti9xx - wss",
				NULL)) {
				snd_port = wss_valid_ports[bits];
				break;
			} else
				continue;
		else if (snd_port == wss_valid_ports[bits])
			break;
	if (snd_port == SND_AUTO_PORT)
		return -EBUSY;
	return snd_opti9xx_config_wss_base(bits);
}

int snd_opti9xx_resources_wss_dma1(void)
{
	int error;
	if ((error = snd_register_dma_channel(opti9xx_card,
		"opti9xx - wss",
		snd_dma1, SND_DMA_TYPE_ISA, snd_dma1_size,
		wss_valid_dma1,
		&snd_dma1_ptr)))
		return error;
	snd_dma1 = snd_dma1_ptr->dma;
	return ESUCCESS;
}

int snd_opti9xx_resources_wss_dma2(void)
{
	int error;
	if ((error = snd_register_dma_channel(opti9xx_card,
		"opti9xx - wss",
		snd_dma2, SND_DMA_TYPE_ISA, snd_dma2_size,
		wss_valid_dma2[snd_dma1],
		&snd_dma2_ptr)))
		return error;
	snd_dma2 = snd_dma2_ptr->dma;
	return ESUCCESS;
}

int snd_opti9xx_resources_wss_irq(void)
{
	int error;
	if ((error = snd_register_interrupt(opti9xx_card,
		"opti9xx - wss",
		snd_irq, SND_IRQ_TYPE_ISA, snd_opti9xx_wss_interrupt,
		NULL,
		wss_valid_irqs + ((opti9xx_chip < c930) ? 1 : 0),
		&snd_irq_ptr)))
		return error;
	snd_irq = snd_irq_ptr->irq;
	return ESUCCESS;
}


int snd_opti9xx_register_ad1848(void)
{
	int error;
	snd_unregister_dma_channels(opti9xx_card);
	snd_dma2 = SND_DEFAULT_DMA1;
	if ((error = snd_opti9xx_resources_wss_dma1()))
		return error;
	if ((error = snd_opti9xx_config_wss_dmas_irq()))
		return error;
	snd_pcm = snd_ad1848_new_device(opti9xx_card,
		WSS_BASE_PORT,
		snd_irq_ptr,
		snd_dma1_ptr,
		AD1848_HW_DETECT);
	if (!snd_pcm)
		return -ENODEV;
#ifdef OPTi9XX_DEBUG
	opti9xx_printk("found an AD1848 compatible device.\n");
#endif	/* OPTi9XX_DEBUG */
	if ((error = snd_pcm_register(snd_pcm, 0)))
		return error;
	opti9xx_wss_irq_ptr = snd_ad1848_interrupt;
	return ESUCCESS;
}

int snd_opti9xx_register_cs4231(void)
{
	int error;
	if ((error = snd_opti9xx_resources_wss_dma1()))
		return error;
	if ((error = snd_opti9xx_resources_wss_dma2()))
		return error;
	if ((error = snd_opti9xx_config_wss_dmas_irq()))
		return error;
	snd_pcm = snd_cs4231_new_device(opti9xx_card,
		WSS_BASE_PORT,
		snd_irq_ptr,
		snd_dma1_ptr, snd_dma2_ptr,
		(opti9xx_chip < c930) ? CS4231_HW_DETECT : CS4231_HW_CS4231);
	if (!snd_pcm)
		return -ENODEV;
#ifdef OPTi9XX_DEBUG
	opti9xx_printk("found a CS4231 compatible device.\n");
#endif	/* OPTi9XX_DEBUG */
	if ((error = snd_pcm_register(snd_pcm, 0)))
		return error;
	opti9xx_wss_irq_ptr = snd_cs4231_interrupt;
	return ESUCCESS;
}

int snd_opti9xx_register_mixer(void)
{
	int error;
	if (opti9xx_wss_irq_ptr == snd_ad1848_interrupt)
		snd_mixer = snd_ad1848_new_mixer(snd_pcm);
	else
		snd_mixer = snd_cs4231_new_mixer(snd_pcm);
	if (!snd_mixer)
		return -ENXIO;
	if ((error =snd_mixer_register(snd_mixer, 0)))
		return error;
	return ESUCCESS;
}

int snd_opti9xx_register_mpu401(void)
{
	int error;
	if (opti9xx_chip < c929)
		return -ENODEV;
	if ((error = snd_opti9xx_resources_mpu401()))
		return error;
	snd_midi = snd_mpu401_uart_new_device(opti9xx_card,
		MPU401_HW_MPU401,
		snd_mpu_port,
		snd_mpu_irq);
	if (!snd_midi)
		return -ENODEV;
	if ((error = snd_rawmidi_register(snd_midi, 0)))
		return error;
#ifdef OPTi9XX_DEBUG
	opti9xx_printk("found a MPU-401 compatible device.\n");
#endif	/* OPTi9XX_DEBUG */
	return ESUCCESS;
}

int snd_opti9xx_register_opl3(void)
{
	int error;
	if ((error = snd_opti9xx_resources_opl3()))
		return error;
	snd_opl3 = snd_opl3_new_device(opti9xx_card,
		OPTi9XX_OPL3_LPORT,
		OPTi9XX_OPL3_RPORT,
		OPL3_HW_OPL3, 0);
	if (!snd_opl3)
		return -ENODEV;
	if ((error = snd_synth_register(snd_opl3)))
		return error;
#ifdef OPTi9XX_DEBUG
	opti9xx_printk("found an OPL3 compatible device.\n");
#endif	/* OPTi9XX_DEBUG */
	return ESUCCESS;
}

int snd_opti9xx_register_wss(void)
{
	int error;
	if ((error = snd_opti9xx_resources_wss_base()))
		return error;
	if ((error = snd_opti9xx_resources_wss_irq()))
		return error;
	if (snd_opti9xx_register_cs4231())
		return snd_opti9xx_register_ad1848();
	return ESUCCESS;
}

int snd_opti9xx_register(void)
{
	int error;
	if ((error = snd_opti9xx_register_wss()))
		return error;
	if ((error = snd_opti9xx_register_mixer()))
		return error;
	if ((error = snd_opti9xx_register_opl3()))
		/* return error; */	/* not needed */
		;
	if ((error = snd_opti9xx_register_mpu401()))
		/* return error; */	/* not needed */
		;
	strcpy(opti9xx_card->abbreviation, OPTi9XX_CHIP);
	sprintf(opti9xx_card->shortname,
		"OPTi %s based soundcard", OPTi9XX_CHIP);
	sprintf(opti9xx_card->longname,
		"OPTi %s based soundcard, WSS at 0x%x, irq %i, dma %i,%i",
		OPTi9XX_CHIP, WSS_BASE_PORT, snd_irq, snd_dma1, snd_dma2);
	opti9xx_card->type=SND_CARD_TYPE_OPTI9XX;
	if (snd_card_register(opti9xx_card))
		return -ENOMEM;
#ifdef OPTi9XX_DEBUG
	opti9xx_printk("WSS at 0x%x, irq %i, dma %i",
		WSS_BASE_PORT, snd_irq, snd_dma1);
	if (snd_dma2 != SND_DEFAULT_DMA1)
		printk(",%i", snd_dma2);
	printk("\n");
	snd_opti9xx_dump();
#endif	/* OPTi9XX_DEBUG */
	return ESUCCESS;
}


int snd_opti9xx_detect_isa(void)
{
	for (opti9xx_chip = c928; opti9xx_chip < unsupported; opti9xx_chip++) {
		unsigned char value, check;
		if (snd_opti9xx_resources_mc_regs())
			continue;
		if (opti9xx_chip < c930) {
			value = snd_opti9xx_read(OPTi9XX_MC(1));
			check = inb(OPTi9XX_MC_BASE + OPTi9XX_MC(1));
			if ((value != 0xff) && (value != check))
				if (value == snd_opti9xx_read(OPTi9XX_MC(1)))
					break;
		}
		else {
			if (snd_opti93x_resources_direct_regs())
				continue;
			outb(OPTi9XX_PASSWD, OPTi9XX_PWD_REG);
			outb(OPTi93X_ENABLE_PWD | OPTi93X_IDX_BITS,
				OPTi9XX_MC_BASE);
			value = snd_opti9xx_read(OPTi9XX_MC(7));
			check = ~value;		/* value ^ 0xff */
			snd_opti9xx_write(OPTi9XX_MC(7), check);
			if (snd_opti9xx_read(OPTi9XX_MC(7)) == check)
				break;
		}
		snd_unregister_ioports(opti9xx_card);
	}
#ifdef OPTi9XX_DEBUG
	if (opti9xx_chip != unsupported) {
		opti9xx_printk("found an OPTi %s chip (ISA)\n", OPTi9XX_CHIP);
		snd_opti9xx_dump();
	}
#endif	/* OPTi9XX_DEBUG */
	return (opti9xx_chip == unsupported) ? -ENODEV : ESUCCESS;
}

int snd_opti9xx_detect_pnp(void)
{
	if (!isapnp_present()) {
		opti9xx_printk("PnP support not present.\n");
		return -ENOSYS;
	}
	return -ENOSYS;
	return ESUCCESS;
}


int snd_opti9xx_probe(void)
{
	int error;
	opti9xx_card = snd_card_new(snd_index, snd_id,
		snd_opti9xx_use_inc, snd_opti9xx_use_dec);
	if (!opti9xx_card)
		return -ENOMEM;
	if (!(error = snd_opti9xx_detect_pnp()))
		error = snd_opti9xx_config_pnp();
	else if (!(error = snd_opti9xx_detect_isa()))
		error = snd_opti9xx_config_isa();
	if (!error)
		return snd_opti9xx_register();
	return error;
}


int snd_opti9xx_check_options(void) {
	int i=0;
	if (snd_port != SND_AUTO_PORT)
		for (; wss_valid_ports[i] != SND_AUTO_PORT; i++)
			if (wss_valid_ports[i] == snd_port)
				break;
	if (wss_valid_ports[i] == SND_AUTO_PORT) {
#ifndef OPTi9XX_EXPERIMENTAL
		return -EINVAL;
#else
		opti9xx_printk("value 0x%x is not valid for snd_port, %s\n",
			snd_port, "changing to 'auto'.");
		snd_port = SND_AUTO_PORT;
#endif	/* OPTi9XX_EXPERIMENTAL */
	}
	if (snd_mpu_port != SND_AUTO_PORT)
		for (i = 0; mpu_valid_ports[i] != SND_AUTO_PORT; i++)
			if (mpu_valid_ports[i] == snd_mpu_port)
				break;
	if (mpu_valid_ports[i] == SND_AUTO_PORT) {
#ifndef OPTi9XX_EXPERIMENTAL
		return -EINVAL;
#else
		opti9xx_printk("value 0x%x is not valid for snd_mpu_port, %s\n",
			snd_port, "changing to 'auto'.");
		snd_mpu_port = SND_AUTO_PORT;
#endif	/* OPTi9XX_EXPERIMENTAL */
	}
	if (snd_mpu_port == cd_port) {
		opti9xx_printk("MPU-401/cdrom port conflict (0x%x): %s\n",
			snd_mpu_port, "disabling cdrom i/f");
		cd_if_sel = 0x00;	/* cdrom drive i/f disabled */
	}
	return ESUCCESS;
}

int snd_opti9xx_parse_options(void) {
	int i;
#ifdef OPTi9XX_DEBUG
	opti9xx_printk("cdrom: %s i/f, port 0x%x, irq %d, drq %d\n",
		cd_if, cd_port, cd_irq, cd_drq);
	opti9xx_printk("game port %s\n", game);
#endif	/* OPTi9XX_DEBUG */
	for (i = 0; i < 6; i++)
		if (!strcmp(cd_if_table[i].name, cd_if))
			break;
	if (i == 6)
		return -EINVAL;
	cd_if_sel = cd_if_table[i].bits;
	for (i = 0; i < 4; i++)
		if (cd_port_table[i].port == cd_port)
			break;
	if (i == 4)
		return -EINVAL;
	cd_port_sel = cd_port_table[i].bits;
	for (i = 0; i < 6; i++)
		if (cd_irq_table[i].irq == cd_irq)
			break;
	if (i == 6)
		return -EINVAL;
	cd_irq_sel = cd_irq_table[i].bits;
	for (i = 0; i < 4; i++)
		if (cd_drq_table[i].drq == cd_drq)
			break;
	if (i == 4)
		return -EINVAL;
	cd_drq_sel = cd_drq_table[i].bits;
	for (i = 0; i < 2; i++)
		if (!strcmp(game_en_table[i].name, game))
			break;
	if (i == 2)
		return -EINVAL;
	game_sel = game_en_table[i].bits;
	return snd_opti9xx_check_options();
}


void cleanup_module(void) {
	if (snd_midi) {
		snd_rawmidi_unregister(snd_midi);
		snd_rawmidi_free(snd_midi);
	}
	if (snd_opl3) {
		snd_synth_unregister(snd_opl3);
		snd_synth_free(snd_opl3);
	}
	if (snd_mixer) {
		snd_mixer_unregister(snd_mixer);
		snd_mixer_free(snd_mixer);
	}
	if (snd_pcm) {
		snd_pcm_unregister(snd_pcm);
		snd_pcm_free(snd_pcm);
	}
	if (opti9xx_card) {
		snd_card_unregister(opti9xx_card);
		snd_card_free(opti9xx_card);
	}
}

int init_module(void) {
	int error;
	if ((error = snd_opti9xx_parse_options()))
		opti9xx_printk("please check module options\n");
	else if ((error = snd_opti9xx_probe()))
		 cleanup_module();
	return error;
}

